/*
 * 	util.h
*/


#ifndef _UTIL_H_
#define _UTIL_H_

#include <stdlib.h>
#include <syslog.h>
#include <netinet/in.h>

#ifdef _UTIL_C_
#define EXTERN
#else
#define EXTERN extern
#endif


#include "setup.h"

/*
 *	useful macros
*/

#define SIN_TO_UINT32(x)	(*((uint32_t *) &(x)))
#define UINT32_TO_SIN(x)	(*((struct in_addr *) &(x)))

#define TERMINATE_STRING(s) 	s[sizeof(s)-1] = '\0';
#define EMPTY_STRING(x) 	(x == NULL || x[0] == '\0')

#define SAFE_CLOSE(x)		while (close(x) == -1 && errno == EINTR);

/*
 *	action()
*/
#define LOG_ALWAYS		0x10

// "rozszerzenie" sysloga na potrzeby programu
// action() przycina priorytet komunikatow do
// LOG_DEBUG jesli sa wyzsze
#define LOG_VERBOSE		0x08


/*
 *	Connection status returned by get_line() callbacks
*/
typedef enum {
	LINE_OK = 0, LINE_CLOSED, LINE_BINARY,		// ---- get_line callback
	LINE_EINTR,					// ---- get_line -------- EINTR=timeout
	LINE_CLOSED_SERVER, LINE_CLOSED_TIMEOUT		// loop -------- --------
} line_status;


typedef line_status (*line_callback)(char* buf, char* pos, int size, void *ptr);


EXTERN int connect_host(struct sockaddr_in dst, int dst_port, struct sockaddr_in src);
EXTERN int setup_listen(char *listen_ip, int listen_port, int backlog);

EXTERN line_status fdgetline(int fd, char* buf, size_t sizeof_buf, int *buf_size, line_callback callback, void *ptr, int *rx_bytes);
EXTERN int fdprintf(int fd, char* format, ...);

EXTERN void action(int prio, char* format, ...);

EXTERN void untaint(/*@null@*/unsigned char *str, int len);
EXTERN void untaint_for_filename(/*@null@*/unsigned char *str, int len);
EXTERN int is_load_above(double max, double *current);
EXTERN char* time2str(time_t t);

#ifdef USE_SHARED_MEM
EXTERN void* shmalloc(size_t size, int *id);
EXTERN int shmfreeid(int id);
#endif



#undef EXTERN

#endif

